@@ -10,6 +10,8 @@
 #include <aws/io/retry_strategy.h>
 #include <aws/s3/s3.h>
 
+AWS_PUSH_SANE_WARNING_LEVEL
+
 struct aws_allocator;
 
 struct aws_http_stream;
@@ -26,6 +28,8 @@ struct aws_s3_meta_request_resume_token;
 struct aws_uri;
 struct aws_string;
 
+struct aws_s3_request_metrics;
+
 /**
  * A Meta Request represents a group of generated requests that are being done on behalf of the
  * original request. For example, one large GetObject request can be transformed into a series
@@ -50,6 +54,16 @@ enum aws_s3_meta_request_type {
     /**
      * The PutObject request will be split into MultiPart uploads that are executed in parallel
      * to improve throughput, when possible.
+     * Note: put object supports both known and unknown body length. The client
+     * relies on Content-Length header to determine length of the body.
+     * Request with unknown content length are always sent using multipart
+     * upload regardless of final number of parts and do have the following limitations:
+     * - multipart threshold is ignored and all request are made through mpu,
+     *   even if they only need one part
+     * - pause/resume is not supported
+     * - meta request will throw error if checksum header is provider (due to
+     *   general limitation of checksum not being usable if meta request is
+     *   getting split)
      */
     AWS_S3_META_REQUEST_TYPE_PUT_OBJECT,
 
@@ -58,18 +72,47 @@ enum aws_s3_meta_request_type {
      * using multiple S3 UploadPartCopy requests in parallel, or bypasses
      * a CopyObject request to S3 if the object size is not large enough for
      * a multipart upload.
+     * Note: copy support is still in development and has following limitations:
+     * - host header must use virtual host addressing style (path style is not
+     *   supported) and both source and dest buckets must have dns compliant name
+     * - only {bucket}/{key} format is supported for source and passing arn as
+     *   source will not work
+     * - source bucket is assumed to be in the same region as dest
      */
     AWS_S3_META_REQUEST_TYPE_COPY_OBJECT,
 
     AWS_S3_META_REQUEST_TYPE_MAX,
 };
 
+/**
+ * The type of S3 request made. Used by metrics.
+ */
+enum aws_s3_request_type {
+    /* Same as the original HTTP request passed to aws_s3_meta_request_options */
+    AWS_S3_REQUEST_TYPE_DEFAULT,
+
+    /* S3 APIs */
+    AWS_S3_REQUEST_TYPE_HEAD_OBJECT,
+    AWS_S3_REQUEST_TYPE_GET_OBJECT,
+    AWS_S3_REQUEST_TYPE_LIST_PARTS,
+    AWS_S3_REQUEST_TYPE_CREATE_MULTIPART_UPLOAD,
+    AWS_S3_REQUEST_TYPE_UPLOAD_PART,
+    AWS_S3_REQUEST_TYPE_ABORT_MULTIPART_UPLOAD,
+    AWS_S3_REQUEST_TYPE_COMPLETE_MULTIPART_UPLOAD,
+    AWS_S3_REQUEST_TYPE_UPLOAD_PART_COPY,
+
+    AWS_S3_REQUEST_TYPE_MAX,
+};
+
 /**
  * Invoked to provide response headers received during execution of the meta request, both for
  * success and error HTTP status codes.
  *
  * Return AWS_OP_SUCCESS to continue processing the request.
- * Return AWS_OP_ERR to indicate failure and cancel the request.
+ *
+ * Return aws_raise_error(E) to cancel the request.
+ * The error you raise will be reflected in `aws_s3_meta_request_result.error_code`.
+ * If you're not sure which error to raise, use AWS_ERROR_S3_CANCELED.
  */
 typedef int(aws_s3_meta_request_headers_callback_fn)(
     struct aws_s3_meta_request *meta_request,
@@ -93,7 +136,10 @@ typedef int(aws_s3_meta_request_headers_callback_fn)(
  * No back-pressure is applied and data arrives as fast as possible.
  *
  * Return AWS_OP_SUCCESS to continue processing the request.
- * Return AWS_OP_ERR to indicate failure and cancel the request.
+ *
+ * Return aws_raise_error(E) to cancel the request.
+ * The error you raise will be reflected in `aws_s3_meta_request_result.error_code`.
+ * If you're not sure which error to raise, use AWS_ERROR_S3_CANCELED.
  */
 typedef int(aws_s3_meta_request_receive_body_callback_fn)(
 
@@ -138,6 +184,17 @@ typedef void(aws_s3_meta_request_progress_fn)(
     const struct aws_s3_meta_request_progress *progress,
     void *user_data);
 
+/**
+ * Invoked to report the telemetry of the meta request once a single request finishes. Invoked from the thread of the
+ * connection that request made from.
+ * Note: *metrics is only valid for the duration of the callback. If you need to keep it around, use
+ * `aws_s3_request_metrics_acquire`
+ */
+typedef void(aws_s3_meta_request_telemetry_fn)(
+    struct aws_s3_meta_request *meta_request,
+    struct aws_s3_request_metrics *metrics,
+    void *user_data);
+
 typedef void(aws_s3_meta_request_shutdown_fn)(void *user_data);
 
 typedef void(aws_s3_client_shutdown_complete_callback_fn)(void *user_data);
@@ -168,6 +225,62 @@ enum aws_s3_checksum_location {
     AWS_SCL_TRAILER,
 };
 
+/**
+ * Info about a single part, for you to review before the upload completes.
+ */
+struct aws_s3_upload_part_review {
+    /* Size in bytes of this part */
+    uint64_t size;
+
+    /* Checksum string, as sent in the UploadPart request (usually base64-encoded):
+     * https://docs.aws.amazon.com/AmazonS3/latest/API/API_UploadPart.html#API_UploadPart_RequestSyntax
+     * This is empty if no checksum is used. */
+    struct aws_byte_cursor checksum;
+};
+
+/**
+ * Info for you to review before an upload completes.
+ *
+ * WARNING: This feature is experimental/unstable.
+ * At this time, review is only available for multipart upload
+ * (when Content-Length is above the `multipart_upload_threshold`,
+ * or Content-Length not specified).
+ */
+struct aws_s3_upload_review {
+    /* The checksum algorithm used. */
+    enum aws_s3_checksum_algorithm checksum_algorithm;
+
+    /* Number of parts uploaded. */
+    size_t part_count;
+
+    /* Array of info about each part uploaded (array is `part_count` in length) */
+    struct aws_s3_upload_part_review *part_array;
+};
+
+/**
+ * Optional callback, for you to review an upload before it completes.
+ * For example, you can review each part's checksum and fail the upload if
+ * you do not agree with them.
+ *
+ * @param meta_request pointer to the aws_s3_meta_request of the upload.
+ * @param info Detailed info about the upload.
+ *
+ * Return AWS_OP_SUCCESS to continue processing the request.
+ *
+ * Return aws_raise_error(E) to cancel the request.
+ * The error you raise will be reflected in `aws_s3_meta_request_result.error_code`.
+ * If you're not sure which error to raise, use AWS_ERROR_S3_CANCELED.
+ *
+ * WARNING: This feature is experimental/unstable.
+ * At this time, the callback is only invoked for multipart upload
+ * (when Content-Length is above the `multipart_upload_threshold`,
+ * or Content-Length not specified).
+ */
+typedef int(aws_s3_meta_request_upload_review_fn)(
+    struct aws_s3_meta_request *meta_request,
+    const struct aws_s3_upload_review *review,
+    void *user_data);
+
 /* Keepalive properties are TCP only.
  * If interval or timeout are zero, then default values are used.
  */
@@ -211,10 +324,18 @@ struct aws_s3_client_config {
     struct aws_signing_config_aws *signing_config;
 
     /* Size of parts the files will be downloaded or uploaded in. */
-    size_t part_size;
+    uint64_t part_size;
 
-    /* If the part size needs to be adjusted for service limits, this is the maximum size it will be adjusted to.. */
-    size_t max_part_size;
+    /* If the part size needs to be adjusted for service limits, this is the maximum size it will be adjusted to. On 32
+     * bit machine, it will be forced to SIZE_MAX, which is around 4GiB. The server limit is 5GiB, but object size limit
+     * is 5TiB for now. We should be good enough for all the cases. */
+    uint64_t max_part_size;
+
+    /* The size threshold in bytes for when to use multipart uploads for a AWS_S3_META_REQUEST_TYPE_PUT_OBJECT meta
+     * request. Uploads over this size will automatically use a multipart upload strategy,while uploads smaller or
+     * equal to this threshold will use a single request to upload the whole object. If not set, `part_size` will be
+     * used as threshold. */
+    uint64_t multipart_upload_threshold;
 
     /* Throughput target in Gbps that we are trying to reach. */
     double throughput_target_gbps;
@@ -236,6 +357,8 @@ struct aws_s3_client_config {
     /**
      * Optional.
      * Proxy configuration for http connection.
+     * If the connection_type is AWS_HPCT_HTTP_LEGACY, it will be converted to AWS_HPCT_HTTP_TUNNEL if tls_mode is
+     * ENABLED. Otherwise, it will be converted to AWS_HPCT_HTTP_FORWARD.
      */
     struct aws_http_proxy_options *proxy_options;
 
@@ -264,6 +387,7 @@ struct aws_s3_client_config {
      * Optional.
      * Configuration options for connection monitoring.
      * If the transfer speed falls below the specified minimum_throughput_bytes_per_second, the operation is aborted.
+     * If set to NULL, default values are used.
      */
     struct aws_http_connection_monitoring_options *monitoring_options;
 
@@ -313,7 +437,7 @@ struct aws_s3_checksum_config {
     enum aws_s3_checksum_algorithm checksum_algorithm;
 
     /**
-     * Enable checksum mode header will be attached to get requests, this will tell s3 to send back checksums headers if
+     * Enable checksum mode header will be attached to GET requests, this will tell s3 to send back checksums headers if
      * they exist. Calculate the corresponding checksum on the response bodies. The meta request will finish with a did
      * validate field and set the error code to AWS_ERROR_S3_RESPONSE_CHECKSUM_MISMATCH if the calculated
      * checksum, and checksum found in the response header do not match.
@@ -336,7 +460,14 @@ struct aws_s3_checksum_config {
     struct aws_array_list *validate_checksum_algorithms;
 };
 
-/* Options for a new meta request, ie, file transfer that will be handled by the high performance client. */
+/**
+ * Options for a new meta request, ie, file transfer that will be handled by the high performance client.
+ *
+ * There are several ways to pass the request's body data:
+ * 1) If the data is already in memory, set the body-stream on `message`.
+ * 2) If the data is on disk, set `send_filepath` for best performance.
+ * 3) If the data will be be produced in asynchronous chunks, set `send_async_stream`.
+ */
 struct aws_s3_meta_request_options {
     /* TODO: The meta request options cannot control the request to be split or not. Should consider to add one */
 
@@ -347,9 +478,26 @@ struct aws_s3_meta_request_options {
      * be used. If not NULL, these options will override the client options. */
     const struct aws_signing_config_aws *signing_config;
 
-    /* Initial HTTP message that defines what operation we are doing. */
+    /* Initial HTTP message that defines what operation we are doing.
+     * Do not set the message's body-stream if the body is being passed by other means (see note above) */
     struct aws_http_message *message;
 
+    /**
+     * Optional.
+     * If set, this file is sent as the request body.
+     * This gives the best performance when sending data from a file.
+     * Do not set if the body is being passed by other means (see note above).
+     */
+    struct aws_byte_cursor send_filepath;
+
+    /**
+     * Optional - EXPERIMENTAL/UNSTABLE
+     * If set, the request body comes from this async stream.
+     * Use this when outgoing data will be produced in asynchronous chunks.
+     * Do not set if the body is being passed by other means (see note above).
+     */
+    struct aws_async_input_stream *send_async_stream;
+
     /**
      * Optional.
      * if set, the flexible checksum will be performed by client based on the config.
@@ -386,17 +534,41 @@ struct aws_s3_meta_request_options {
 
     /**
      * Invoked to report progress of the meta request execution.
-     * Currently, the progress callback is invoked only for the CopyObject meta request type.
-     * TODO: support this callback for all the types of meta requests
-     * See `aws_s3_meta_request_progress_fn`
      */
     aws_s3_meta_request_progress_fn *progress_callback;
 
     /**
      * Optional.
-     * The endpoint for the meta request to connect to. If not set, then tls settings from client will
-     * determine the port, and host header from `message` will determine the host for the connection.
-     * Note: must match the host header from `message`
+     * To get telemetry metrics when a single request finishes.
+     * If set the request will keep track of the metrics from `aws_s3_request_metrics`, and fire the callback when the
+     * request finishes receiving response.
+     * See `aws_s3_meta_request_telemetry_fn`
+     *
+     * Notes:
+     * - The callback will be invoked multiple times from different threads.
+     */
+    aws_s3_meta_request_telemetry_fn *telemetry_callback;
+
+    /**
+     * Optional.
+     * Callback for reviewing an upload before it completes.
+     * WARNING: experimental/unstable
+     * See `aws_s3_upload_review_fn`
+     */
+    aws_s3_meta_request_upload_review_fn *upload_review_callback;
+
+    /**
+     * Optional.
+     * Endpoint override for request. Can be used to override scheme and port of
+     * the endpoint.
+     * There is some overlap between Host header and Endpoint and corner cases
+     * are handled as follows:
+     * - Only Host header is set - Host is used to construct endpoint. https is
+     *   default with corresponding port
+     * - Only endpoint is set - Host header is created from endpoint. Port and
+     *   Scheme from endpoint is used.
+     * - Both Host and Endpoint is set - Host header must match Authority of
+     *   Endpoint uri. Port and Scheme from endpoint is used.
      */
     struct aws_uri *endpoint;
 
@@ -425,7 +597,7 @@ struct aws_s3_meta_request_result {
     /* HTTP Headers for the failed request that triggered finish of the meta request.  NULL if no request failed. */
     struct aws_http_headers *error_response_headers;
 
-    /* Response body for the failed request that triggered finishing of the meta request.  NUll if no request failed.*/
+    /* Response body for the failed request that triggered finishing of the meta request.  NULL if no request failed.*/
     struct aws_byte_buf *error_response_body;
 
     /* Response status of the failed request or of the entire meta request. */
@@ -437,7 +609,7 @@ struct aws_s3_meta_request_result {
      * uploaded as a multipart object.
      *
      * If the object to get is multipart object, the part checksum MAY be validated if the part size to get matches the
-     * part size uploaded. In that case, if any part mismatch the checksum received, the meta request will failed with
+     * part size uploaded. In that case, if any part mismatch the checksum received, the meta request will fail with
      * checksum mismatch. However, even if the parts checksum were validated, this will NOT be set to true, as the
      * checksum for the whole meta request was NOT validated.
      **/
@@ -543,11 +715,13 @@ int aws_s3_meta_request_pause(
  * to persist those in whichever way they choose.
  */
 struct aws_s3_upload_resume_token_options {
-    struct aws_byte_cursor upload_id;
-    size_t part_size;
-    size_t total_num_parts;
+    struct aws_byte_cursor upload_id; /* Required */
+    uint64_t part_size;               /* Required. Must be less than SIZE_MAX */
+    size_t total_num_parts;           /* Required */
 
-    /*
+    /**
+     * Optional.
+     *
      * Note: during resume num_parts_uploaded is used for sanity checking against
      * uploads on s3 side.
      * In cases where upload id does not exist (already resumed using this token
@@ -591,7 +765,7 @@ enum aws_s3_meta_request_type aws_s3_meta_request_resume_token_type(
  * Part size associated with operation.
  */
 AWS_S3_API
-size_t aws_s3_meta_request_resume_token_part_size(struct aws_s3_meta_request_resume_token *resume_token);
+uint64_t aws_s3_meta_request_resume_token_part_size(struct aws_s3_meta_request_resume_token *resume_token);
 
 /*
  * Total num parts associated with operation.
@@ -632,12 +806,177 @@ struct aws_s3_meta_request *aws_s3_meta_request_acquire(struct aws_s3_meta_reque
 AWS_S3_API
 struct aws_s3_meta_request *aws_s3_meta_request_release(struct aws_s3_meta_request *meta_request);
 
+/**
+ * Initialize the configuration for a default S3 signing.
+ */
 AWS_S3_API
 void aws_s3_init_default_signing_config(
     struct aws_signing_config_aws *signing_config,
     const struct aws_byte_cursor region,
     struct aws_credentials_provider *credentials_provider);
 
+/**
+ * Add a reference, keeping this object alive.
+ * The reference must be released when you are done with it, or it's memory will never be cleaned up.
+ * Always returns the same pointer that was passed in.
+ */
+AWS_S3_API
+struct aws_s3_request_metrics *aws_s3_request_metrics_acquire(struct aws_s3_request_metrics *metrics);
+
+/**
+ * Release a reference.
+ * When the reference count drops to 0, this object will be cleaned up.
+ * It's OK to pass in NULL (nothing happens).
+ * Always returns NULL.
+ */
+AWS_S3_API
+struct aws_s3_request_metrics *aws_s3_request_metrics_release(struct aws_s3_request_metrics *metrics);
+
+/*************************************  Getters for s3 request metrics ************************************************/
+/**
+ * Get the request ID from aws_s3_request_metrics.
+ * If unavailable, AWS_ERROR_S3_METRIC_DATA_NOT_AVAILABLE will be raised.
+ * If available, out_request_id will be set to a string. Be warned this string's lifetime is tied to the metrics
+ * object.
+ **/
+AWS_S3_API
+int aws_s3_request_metrics_get_request_id(
+    const struct aws_s3_request_metrics *metrics,
+    const struct aws_string **out_request_id);
+
+/* Get the start time from aws_s3_request_metrics, which is when S3 client prepare the request to be sent. Always
+ * available. Timestamp are from `aws_high_res_clock_get_ticks`  */
+AWS_S3_API
+void aws_s3_request_metrics_get_start_timestamp_ns(
+    const struct aws_s3_request_metrics *metrics,
+    uint64_t *out_start_time);
+
+/* Get the end time from aws_s3_request_metrics. Always available */
+AWS_S3_API
+void aws_s3_request_metrics_get_end_timestamp_ns(const struct aws_s3_request_metrics *metrics, uint64_t *out_end_time);
+
+/* Get the total duration time from aws_s3_request_metrics. Always available */
+AWS_S3_API
+void aws_s3_request_metrics_get_total_duration_ns(
+    const struct aws_s3_request_metrics *metrics,
+    uint64_t *out_total_duration);
+
+/* Get the time stamp when the request started to be encoded. Timestamps are from `aws_high_res_clock_get_ticks`
+ * AWS_ERROR_S3_METRIC_DATA_NOT_AVAILABLE will be raised if the request ended before it gets sent. */
+AWS_S3_API
+int aws_s3_request_metrics_get_send_start_timestamp_ns(
+    const struct aws_s3_request_metrics *metrics,
+    uint64_t *out_send_start_time);
+
+/* Get the time stamp when the request finished to be encoded. Timestamps are from `aws_high_res_clock_get_ticks`
+ * AWS_ERROR_S3_METRIC_DATA_NOT_AVAILABLE will be raised if data not available. */
+AWS_S3_API
+int aws_s3_request_metrics_get_send_end_timestamp_ns(
+    const struct aws_s3_request_metrics *metrics,
+    uint64_t *out_send_end_time);
+
+/* The time duration for the request from start encoding to finish encoding (send_end_timestamp_ns -
+ * send_start_timestamp_ns).
+ * AWS_ERROR_S3_METRIC_DATA_NOT_AVAILABLE will be raised if data not available. */
+AWS_S3_API
+int aws_s3_request_metrics_get_sending_duration_ns(
+    const struct aws_s3_request_metrics *metrics,
+    uint64_t *out_sending_duration);
+
+/* Get the time stamp when the response started to be received from the network channel. Timestamps are from
+ * `aws_high_res_clock_get_ticks` AWS_ERROR_S3_METRIC_DATA_NOT_AVAILABLE will be raised if data not available. */
+AWS_S3_API
+int aws_s3_request_metrics_get_receive_start_timestamp_ns(
+    const struct aws_s3_request_metrics *metrics,
+    uint64_t *out_receive_start_time);
+
+/* Get the time stamp when the response finished to be received from the network channel. Timestamps are from
+ * `aws_high_res_clock_get_ticks` AWS_ERROR_S3_METRIC_DATA_NOT_AVAILABLE will be raised if data not available. */
+AWS_S3_API
+int aws_s3_request_metrics_get_receive_end_timestamp_ns(
+    const struct aws_s3_request_metrics *metrics,
+    uint64_t *out_receive_end_time);
+
+/* The time duration for the request from start receiving to finish receiving (receive_end_timestamp_ns -
+ * receive_start_timestamp_ns).
+ * AWS_ERROR_S3_METRIC_DATA_NOT_AVAILABLE will be raised if data not available. */
+AWS_S3_API
+int aws_s3_request_metrics_get_receiving_duration_ns(
+    const struct aws_s3_request_metrics *metrics,
+    uint64_t *out_receiving_duration);
+
+/* Get the response status code for the request. AWS_ERROR_S3_METRIC_DATA_NOT_AVAILABLE will be raised if data not
+ * available. */
+AWS_S3_API
+int aws_s3_request_metrics_get_response_status_code(
+    const struct aws_s3_request_metrics *metrics,
+    int *out_response_status);
+
+/* Get the HTTP Headers of the response received for the request. AWS_ERROR_S3_METRIC_DATA_NOT_AVAILABLE will be raised
+ * if data not available. */
+AWS_S3_API
+int aws_s3_request_metrics_get_response_headers(
+    const struct aws_s3_request_metrics *metrics,
+    struct aws_http_headers **out_response_headers);
+
+/**
+ * Get the path and query of the request.
+ * If unavailable, AWS_ERROR_S3_METRIC_DATA_NOT_AVAILABLE will be raised.
+ * If available, out_request_path_query will be set to a string. Be warned this string's lifetime is tied to the metrics
+ * object.
+ */
+AWS_S3_API
+void aws_s3_request_metrics_get_request_path_query(
+    const struct aws_s3_request_metrics *metrics,
+    const struct aws_string **out_request_path_query);
+
+/**
+ * Get the host_address of the request.
+ * If unavailable, AWS_ERROR_S3_METRIC_DATA_NOT_AVAILABLE will be raised.
+ * If available, out_host_address will be set to a string. Be warned this string's lifetime is tied to the metrics
+ * object.
+ */
+AWS_S3_API
+void aws_s3_request_metrics_get_host_address(
+    const struct aws_s3_request_metrics *metrics,
+    const struct aws_string **out_host_address);
+
+/**
+ * Get the IP address of the request connected to.
+ * If unavailable, AWS_ERROR_S3_METRIC_DATA_NOT_AVAILABLE will be raised.
+ * If available, out_ip_address will be set to a string. Be warned this string's lifetime is tied to the metrics object.
+ */
+AWS_S3_API
+int aws_s3_request_metrics_get_ip_address(
+    const struct aws_s3_request_metrics *metrics,
+    const struct aws_string **out_ip_address);
+
+/* Get the id of connection that request was made from. AWS_ERROR_S3_METRIC_DATA_NOT_AVAILABLE will be raised if data
+ * not available */
+AWS_S3_API
+int aws_s3_request_metrics_get_connection_id(const struct aws_s3_request_metrics *metrics, size_t *out_connection_id);
+
+/* Get the thread ID of the thread that request was made from. AWS_ERROR_S3_METRIC_DATA_NOT_AVAILABLE will be raised if
+ * data not available */
+AWS_S3_API
+int aws_s3_request_metrics_get_thread_id(const struct aws_s3_request_metrics *metrics, aws_thread_id_t *out_thread_id);
+
+/* Get the stream-id, which is the idex when the stream was activated. AWS_ERROR_S3_METRIC_DATA_NOT_AVAILABLE will be
+ * raised if data not available */
+AWS_S3_API
+int aws_s3_request_metrics_get_request_stream_id(const struct aws_s3_request_metrics *metrics, uint32_t *out_stream_id);
+
+/* Get the request type from request metrics. */
+AWS_S3_API
+void aws_s3_request_metrics_get_request_type(
+    const struct aws_s3_request_metrics *metrics,
+    enum aws_s3_request_type *out_request_type);
+
+/* Get the AWS CRT error code from request metrics. */
+AWS_S3_API
+int aws_s3_request_metrics_get_error_code(const struct aws_s3_request_metrics *metrics);
+
 AWS_EXTERN_C_END
+AWS_POP_SANE_WARNING_LEVEL
 
 #endif /* AWS_S3_CLIENT_H */
